#![allow(dead_code)]

extern crate rand;

// Collection ordonnée de `T`
enum ArBinaire<T> {
    Vide,
    NonVide(Box<Noeud<T>>)
}

// Portion de ArBinaire.
struct Noeud<T> {
    element: T,
    gauche: ArBinaire<T>,
    droite: ArBinaire<T>
}

#[test]
fn binary_tree_size() {
    use std::mem::size_of;

    let word = size_of::<usize>();
    assert_eq!(size_of::<ArBinaire<String>>(), word);
    type Triple = (&'static str, ArBinaire<&'static str>, ArBinaire<&'static str>);
    assert_eq!(size_of::<Triple>(), 4 * word);
}

#[test]
fn test_hand_building_tree_of_planets() {
    use self::ArBinaire::*;
    let jupiter_tree = NonVide(Box::new(Noeud {
        element: "Jupiter",
        gauche: Vide,
        droite: Vide
    }));
    let mercury_tree = NonVide(Box::new(Noeud {
        element: "Mercure",
        gauche: Vide,
        droite: Vide
    }));
    let mars_tree = NonVide(Box::new(Noeud {
        element: "Mars",
        gauche: jupiter_tree,
        droite: mercury_tree
    }));
    let venus_tree = NonVide(Box::new(Noeud {
        element: "Venus",
        gauche: Vide,
        droite: Vide
    }));
    let uranus_tree = NonVide(Box::new(Noeud {
        element: "Uranus",
        gauche: Vide,
        droite: venus_tree
    }));
    let tree = NonVide(Box::new(Noeud {
        element: "Saturne",
        gauche: mars_tree,
        droite: uranus_tree
    }));

    assert_eq!(tree.walk(),
               vec!["Jupiter", "Mars", "Mercure", "Saturne", "Uranus", "Venus"]);
}

impl<T: Clone> ArBinaire<T> {
    fn walk(&self) -> Vec<T> {
        match *self {
            ArBinaire::Vide => vec![],
            ArBinaire::NonVide(ref boxed) => {
                let mut result = boxed.gauche.walk();
                result.push(boxed.element.clone());
                result.extend(boxed.droite.walk());
                result
            }
        }
    }
}

impl<T: Ord> ArBinaire<T> {
    fn add(&mut self, value: T) {
        match *self {
            ArBinaire::Vide =>
                *self = ArBinaire::NonVide(Box::new(Noeud {
                    element: value,
                    gauche: ArBinaire::Vide,
                    droite: ArBinaire::Vide
                })),
            ArBinaire::NonVide(ref mut node) =>
                if value <= node.element {
                    node.gauche.add(value);
                } else {
                    node.droite.add(value);
                }
        }
    }
}

#[test]
fn test_add_method_1() {
    let planets = vec!["Mercure", "Venus", "Mars", "Jupiter", "Saturne", "Uranus"];
    let mut tree = ArBinaire::Vide;
    for planet in planets {
        tree.add(planet);
    }

    assert_eq!(tree.walk(),
               vec!["Jupiter", "Mars", "Mercure", "Saturne", "Uranus", "Venus"]);
}

#[test]
fn test_add_method_2() {
    let mut tree = ArBinaire::Vide;
    tree.add("Mercure");
    tree.add("Venus");
    for planet in vec!["Mars", "Jupiter", "Saturne", "Uranus"]  {
        tree.add(planet);
    }

    assert_eq!(tree.walk(),
               vec!["Jupiter", "Mars", "Mercure", "Saturne", "Uranus", "Venus"]);
}

use self::ArBinaire::*;

// The state of an in-order traversal of a `ArBinaire`.
struct TreeIter<'a, T: 'a> {
    // A stack of references to tree nodes. Since we use `Vec`'s
    // `push` and `pop` methods, the top of the stack is the end of the
    // vector.
    //
    // The node the iterator will visit next is at the top of the stack,
    // with those ancestors still unvisited below it. If the stack is empty,
    // the iteration is over.
    unvisited: Vec<&'a Noeud<T>>
}

impl<'a, T: 'a> TreeIter<'a, T> {
    fn push_gauche_edge(&mut self, mut tree: &'a ArBinaire<T>) {
        while let NonVide(ref node) = *tree {
            self.unvisited.push(node);
            tree = &node.gauche;
        }
    }
}

impl<T> ArBinaire<T> {
    fn iter(&self) -> TreeIter<T> {
        let mut iter = TreeIter { unvisited: Vec::new() };
        iter.push_gauche_edge(self);
        iter
    }
}

impl<'a, T: 'a> IntoIterator for &'a ArBinaire<T> {
    type Item = &'a T;
    type IntoIter = TreeIter<'a, T>;
    fn into_iter(self) -> Self::IntoIter {
        self.iter()
    }
}

impl<'a, T> Iterator for TreeIter<'a, T> {
    type Item = &'a T;
    fn next(&mut self) -> Option<&'a T> {
        // Find the node this iteration must produce,
        // or finish the iteration.
        let node = match self.unvisited.pop() {
            None => return None,
            Some(n) => n
        };

        // The next node after this one is the gauchemost child of
        // this node's droite child, so push the path from here down.
        self.push_gauche_edge(&node.droite);

        // Produce a reference to this node's value.
        Some(&node.element)
    }
}

#[test]
fn external_iterator() {
    fn make_node<T>(gauche: ArBinaire<T>, element: T, droite: ArBinaire<T>)
               -> ArBinaire<T>
    {
        NonVide(Box::new(Noeud { gauche, element, droite }))
    }

    // Build a small tree.
    let subtree_l = make_node(Vide, "mecha", Vide);
    let subtree_rl = make_node(Vide, "droid", Vide);
    let subtree_r = make_node(subtree_rl, "robot", Vide);
    let tree = make_node(subtree_l, "Jaeger", subtree_r);

    // Iterate over it.
    let mut v = Vec::new();
    for kind in &tree {
        v.push(*kind);
    }
    assert_eq!(v, ["mecha", "Jaeger", "droid", "robot"]);

    let gauche_subtree = make_node(Vide, "mecha", Vide);
    let droite_subtree = make_node(make_node(Vide, "droid", Vide),
                                  "robot",
                                  Vide);
    let tree = make_node(gauche_subtree, "Jaeger", droite_subtree);

    let mut v = Vec::new();
    let mut iter = TreeIter { unvisited: vec![] };
    iter.push_gauche_edge(&tree);
    for kind in iter {
        v.push(*kind);
    }
    assert_eq!(v, ["mecha", "Jaeger", "droid", "robot"]);

    let mut v = Vec::new();
    for kind in &tree {
        v.push(*kind);
    }
    assert_eq!(v, ["mecha", "Jaeger", "droid", "robot"]);

    let mut v = Vec::new();
    let mut state = tree.into_iter();
    while let Some(kind) = state.next() {
        v.push(*kind);
    }
    assert_eq!(v, ["mecha", "Jaeger", "droid", "robot"]);

    assert_eq!(tree.iter()
               .map(|name| format!("mega-{}", name))
               .collect::<Vec<_>>(),
               vec!["mega-mecha", "mega-Jaeger",
                    "mega-droid", "mega-robot"]);

    let mut iterator = tree.into_iter();
    assert_eq!(iterator.next(), Some(&"mecha"));
    assert_eq!(iterator.next(), Some(&"Jaeger"));
    assert_eq!(iterator.next(), Some(&"droid"));
    assert_eq!(iterator.next(), Some(&"robot"));
    assert_eq!(iterator.next(), None);
}

#[test]
fn other_cloned() {
    use std::collections::BTreeSet;

    let mut set = BTreeSet::new();
    set.insert("mecha");
    set.insert("Jaeger");
    set.insert("droid");
    set.insert("robot");
    assert_eq!(set.iter().cloned().collect::<Vec<_>>(),
               ["Jaeger", "droid", "mecha", "robot"]);
}

#[test]
fn fuzz() {
    fn make_random_tree(p: f32) -> ArBinaire<i32> {
        use rand::{ThreadRng, thread_rng};
        use rand::distributions::range::Range;
        use rand::distributions::Sample;

        fn make(p: f32, next: &mut i32, rng: &mut ThreadRng) -> ArBinaire<i32> {
            let mut range = Range::new(0.0, 1.0);
            if range.sample(rng) > p {
                Vide
            } else {
                let gauche = make(p * p, next, rng);
                let element = *next;
                *next += 1;
                let droite = make(p * p, next, rng);
                NonVide(Box::new(Noeud { gauche, element, droite }))
            }
        }

        make(p, &mut 0, &mut thread_rng())
    }

    for _ in 0..100 {
        let tree = make_random_tree(0.9999);
        assert!(tree.into_iter().fold(Some(0), |s, &i| {
            s.and_then(|expected| if i == expected { Some(expected+1) } else { None })
        }).is_some());
    }
}
